// evmovie.c

// Programmer: Ken Brakke
//             Math Dept
//             Susquehanna University
//             Selinsgrove, PA 17870
//             brakke@susqu.edu
//             http://www.susqu.edu/brakke

// Reads in COFF files from Evolver output and displays in 
// animated 3D form.

// Compilation:
// Needs opengl and glut libraries.
// In Windows, link with setargv.obj, glut32.lib
// In unix, link with -lGL -lGLU -lglut

// Usage:
//    evmovie filelist
// where filelist is list of frame files in OFF-like format described below.

// Frame file formats:
// ASCII:
/*  line 1: OFF
    with optional parameters on the same line:
         OPACITY     if the "opacity" facet attribute is being used
         BBOX        if the bounding box is to be shown at startup
         LOOP        if the looping movie mode is to be run at startup
         CYCLE       if the cycle movie mode is to be run at startup
         PLAY        if the forward movie mode is to be run at startup
         CLIP        if clip mode is to be enabled at startup
         SMOOTH      if smooth shading is to be enabled at startup
         SHOW_EDGES  if all facet edges are to be shown at startup
         FACETS_OFF  if facets are not to be shown at startup
         VIEW        gives the initial view matrix.  Must be followed
                     by 16 numbers, the components of the view matrix.
                     Can use values put out by the Evolver's
                     "print view_matrix" command.  Example (the default view):
                        VIEW 1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1
         TEXT        gives a string of text to be displayed in the frame.
                     Must be followed by x,y window coordinates for the
                     lower left start of the text, then a font size number
                     (also relative to window size), and a quoted string
                     of text.  Example:
                         TEXT  0.01 0.01 0.03 "Hello World!"
*/
//  line 2: nv nf ne
//    where nv is the number of vertices, 
//          nf is the number of facets
//          ne is the number of edges (yes, edges last)
//  lines 3 to 2+nv:
//    three floating point numbers per line, coordinates of a vertex
//  lines 3+nv to 2+nv+nf:
//    3 v1 v2 v3 c
//    where 3 is the number of vertices in the polygon (from OFF format,
//    but 3 is the only number evmovie recognizes),
//    v1, v2, v3 are the vertex numbers of the facet in the vertex list,
//      using 0 based indexing,
//    c is the color of the facet, an integer from 0 to 15, being the
//      same color number used by Evolver.
//  lines 3+nv+nf to 2+nv+nf+ne:
//    2 v1 v2 c
//    where 2 is the number of vertices in the edge,
//    v1, v2 are the vertex numbers of the edge ends in the vertex list,
//      using 0 based indexing,
//    c is the color of the facet, an integer from 0 to 15, being the
//      same color number used by Evolver.
//    Only special edges that should always be shown should be in this list.

// Binary:
//  line 1 is ascii, ending with newline: OFF BINARY
//  with options on line 1 as in ASCII version.
//  The rest of the file is the same format as the ascii version, but
//  with 4-byte ints and 4-byte floats and no newlines.  So the
//  length of the file after the first line is 3 + nv*4*3 + nf*4*5 + ne*4*4. 
//  Beware of generating frame files on a big-endian machine and
//  running them on a little-endian machine.

// Note: format is like OFF, except each facet has a color map number added.
// (see http://www.geom.uiuc.edu/software/geomview/geomview_6.html#SEC36)

#define _CRT_SECURE_NO_WARNINGS

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#if defined(WIN32) || defined(_WIN32)
#include <glut.h>
#include <windows.h>
#include <sys/stat.h>
#include <io.h>

#else

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#define __cdecl
#define _fileno fileno
int _filelength(int fd)
{ struct stat statbuf;
  fstat(fd,&statbuf);
  return statbuf.st_size;
}
#if defined(__APPLE__)
#include <glut.h>
#else
// unix, linux
#include <GL/glut.h>
#endif

#endif

#define _USE_MATH_DEFINES
#include <string.h>
#include <math.h>


#define GO 1
#define GOBACK 2
#define STOP 3

int go_state = STOP;

#define NOLOOP 100
#define LOOP   101
#define CYCLE   102
int loop_mode = NOLOOP;

#define FORWARD 201
#define BACKWARD 202
int go_direction = FORWARD;

int binary_flag; // for binary input files
int opacity_enable = 1;

int translucency_flag = 0; // for T key, forced translucency

typedef double MAT4[4][4];
void matvec_mul(MAT4 a, double *, double *, int , int );

int ticks_per_frame = CLOCKS_PER_SEC/20; // clock ticks per frame
int msec_per_tick = 1000/CLOCKS_PER_SEC;
int then,now;  // for timing inter-frame interval
void idle_func(void);
void load_frame(int);

// Flag values; UNSET for reading option from file
#define UNSET -1
#define ON 1
#define OFF 0

int edge_show_flag = UNSET;  // for showing all triangulation edges
int facet_show_flag = 1; // for showing all facets
int smooth_flag = UNSET;     // for smooth shading

int xsize=400,ysize=400; // window size in pixels
double scrx[4],scry[4];
double xscale,yscale;
double window_aspect_ratio = 1;
double window_aspect = 1;
double aspect_flag = 0;
double aspect = 1.0;
double imagescale = 1;
double bbox_min[3] = {1e30,1e30,1e30};
double bbox_max[3] = {-1e30,-1e30,-1e30};
int bounding_box_flag = UNSET; // whether to show bounding box
GLfloat linewidth = 1.0f;
double edge_bias = 0.001;

double clip_coeff[5][4] = {{-1,0,0,0}};
double init_clip_coeff[4] = {-1,0,0,0};  // from file options
double world_clip[4];  // in world coordinates
int init_clip_flag;   // set if file has clip plane

int mainmenu;

#define SDIM 3
#define HOMDIM 4
MAT4 view = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
MAT4 from_focus = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
MAT4 to_focus = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
MAT4 identmat = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
MAT4 flip = {{0,0,-1,0},{1,0,0,0},{0,-1,0,0},{0,0,0,1}};
int view_set_flag = 0; // if view matrix in option line of input file
MAT4 init_view = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};

// Color map, copied from Surface Evolver
float rgba[16][4] = {  /* rgba */ {0.0,0.0,0.0,1.},{0.0,0.0,1.,1.},
 {0.0,1.,0.0,1.},{0.0,1.,1.,1.}, {1.,0.0,0.0,1.},{1.,0.0,1.,1.},
 {1.,0.5,0.,1.},{.6,.6,.6,1.},{.3,.3,.3,1.},{.3,.8,1.,1.}, {.5,1.,.5,1.},
 {.5,1.,1.,1.},{1.,.5,.5,1.},{1.,.5,1.,1.},{1.,1.,.0,1.},{1.,1.,1.,1.} };
enum COLORS {
     BLACK,    /* dark colors */
     BLUE,
     GREEN,
     CYAN,
     RED,
     MAGENTA,
     BROWN,
     LIGHTGRAY,
     DARKGRAY, /* light colors */
     LIGHTBLUE,
     LIGHTGREEN,
     LIGHTCYAN,
     LIGHTRED,
     LIGHTMAGENTA,
     YELLOW,
     WHITE
};
int background_color = LIGHTBLUE;

 /* lighting info for surface */
 static GLfloat mat_specular[] = {.5f,.5f,.5f,1.0f};
 static GLfloat mat_shininess[] = {10.0f};
 static GLfloat mat_diffuse[] = {1.0f,1.0f,1.0f,1.0f};
 static GLfloat mat_white[] = {1.0f,1.0f,1.0f,1.0f};
 static GLfloat mat_emission[] = {0.3f,0.3f,0.3f,1.0f};
#define INTENSITY1  0.5f
 static GLfloat light0_position[] = {1.0f,0.0f,1.0f,0.0f};  /* front */
 static GLfloat light0_diffuse[] = {INTENSITY1,INTENSITY1,INTENSITY1,1.0f};
 static GLfloat light0_ambient[] = {.3f,.3f,.3f,1.0f};

#define INTENSITY2  0.5f
 static GLfloat light1_position[] = {0.0f,0.0f,1.0f,0.0f};  /* above */
 static GLfloat light1_diffuse[] = {INTENSITY2,INTENSITY2,INTENSITY2,1.0f};
 static GLfloat light_none[] = {0.f,0.f,0.f,.0f};

#define MAXSTRING 200
#define MAXTEXT 10
struct text_t { double x,y;
                double scale;
                char string[MAXSTRING+2];
};

// Data structure for each movie frame.
struct frame_t {
  int loaded_flag; // whether data is loaded
  int endian_swap_flag; // whether data in wrong endian form
  int opacity_flag;
  int bbox_flag;
  int vcount; 
  int fcount;
  float *vertexdata;
  float *normaldata;
  int *facetdata; 
  float *colordata;
  int ecount;
  int ec_counts[16];
  int *edgedata;  // edges grouped by color
  int timestamp;
  int filelength; // size of file on disk
  struct text_t *text[MAXTEXT];
} *frames;
int frame_count;
int this_frame;
double available_bytes; // to see how many frames can fit in memory at once
void **free_memory_frames;  // free list of memory blocks for frames
int free_memory_count;  // number of frames in list
int global_timestamp;  // for frame use and LRU replacement

char *file_buffer; // for reading whole file at once
#define MAXFILES 100000
char *filenames[MAXFILES]; // list of image file names
int maxmem;  // maximum data memory size
int maxfilesize;  // maximum binary file size
int maxfacets; // maximum number of facets in any frame

int oldx,oldy,mouse_left_state = GLUT_UP;
int newx,newy,basex,basey;
// Mouse states
#define MM_ROTATE       1
#define MM_TRANSLATE   2
#define MM_SCALE       3
#define MM_SPIN        4
#define MM_SLICE       5
#define MM_SLICE_SPIN  6
// modes 1-6 remembered in last_mouse_mode while higher ones in effect
#define MM_SLIDER      7
#define MM_GO_BUTTON   8
#define MM_LEFT_ARROW  9
#define MM_RIGHT_ARROW 10
int mouse_mode = MM_ROTATE;
int last_mouse_mode = MM_ROTATE; // remember during sliding 
int clip_view_flag = 0; 

void read_file_header_binary(char *,int );
void read_file_binary(char *, int );

// For handling wrong-endian binary files
void endian_swap(void *a,int size)
{ unsigned char *aa = a;
  unsigned char tmp;
  int n;
  for ( n = 0 ; n < size/2 ; n++ )
  { tmp = aa[n];
    aa[n] = aa[size-n-1];
    aa[size-n-1] = tmp;
  }
}
  
/******************************************************************
* Function: calc_normals()
*
* Purpose: Calculate normals to vertices for smooth shading.
*          Rather crude algorithm that assumes all facets 
*          consistently oriented, and just averages all normals
*          of facets adjacent to vertex.
*/
void calc_normals(struct frame_t *f)
{ int i,j,k;

  // Calculate normals
  for ( i = 0 ; i < f->fcount ; i++ )
  { float *t,*h,*hh,nx[3];
    t = f->vertexdata + 3*f->facetdata[3*i];
    h = f->vertexdata + 3*f->facetdata[3*i+1];
    hh = f->vertexdata + 3*f->facetdata[3*i+2];
    nx[0] = (h[1]-t[1])*(hh[2]-t[2]) - (h[2]-t[2])*(hh[1]-t[1]);
    nx[1] = (h[2]-t[2])*(hh[0]-t[0]) - (h[0]-t[0])*(hh[2]-t[2]);
    nx[2] = (h[0]-t[0])*(hh[1]-t[1]) - (h[1]-t[1])*(hh[0]-t[0]);
    for ( j = 0 ; j < 3 ; j++ )
      for ( k = 0  ; k < 3 ; k++ )
        f->normaldata[3*f->facetdata[3*i+j]+k] += nx[k];
 }
 for ( i = 0 ; i < f->vcount ; i++ )
 { double mag;
   mag = f->normaldata[3*i]*f->normaldata[3*i] + 
          f->normaldata[3*i+1]*f->normaldata[3*i+1] + 
          f->normaldata[3*i+2]*f->normaldata[3*i+2];
   mag = 1/sqrt(mag);
   for ( j = 0 ; j < 3 ; j++ )
       f->normaldata[3*i+j] *= mag;
 }
}  // enc calc_normals()

/****************************************************************
* Function: ec_sort()
* Purpose: Used by qsort() in color_sort() below to sort
*          edges into color groups. Inputs are pointers
*          to implicit structures with integer fields v1,v2,v3,c
*/
int ec_sort(const void *aa, const void *bb)
{ int *a = (int*)aa;
  int *b = (int*)bb;
  if (a[2] < b[2]) return -1;
  if (a[2] > b[2]) return  1;
  return 0;
} // end ec_sort()

  
void color_sort(struct frame_t *f)
{ int i,j;

  // Sort edges by color
  qsort(f->edgedata,f->ecount,3*sizeof(int),ec_sort);
  
  // Count each color
  for ( i = 0 ; i < f->ecount ; i++ )
    f->ec_counts[f->edgedata[3*i+2]]++;
    
  // Strip individual color info, so vertex indices contiguous
  for ( i = 0 ; i < f->ecount ; i++ )
    for ( j = 0 ; j < 2 ; j++ )
      f->edgedata[2*i+j] = f->edgedata[3*i+j];
      
  
} // end color_sort

// Read various options embedded in first line
void read_options(int frame_number, char *line)
{ struct frame_t *f = frames+frame_number;
  char *spot;
  int textid;
  int retval;

  if ( strstr(line,"OPACITY") )
    f->opacity_flag = 1;
  else f->opacity_flag = 0;

  if ( strstr(line,"BBOX") && bounding_box_flag == UNSET )
    bounding_box_flag = ON;

  if ( strstr(line,"LOOP") && frame_number==0 )
  { loop_mode = LOOP;
    go_state = GO;
  }

  if ( strstr(line,"CYCLE") && frame_number==0 )
  { loop_mode = CYCLE;
    go_state = GO;
  }

  if ( strstr(line,"PLAY") && frame_number==0 )
  { loop_mode = NOLOOP;
    go_state = GO;
  }

  if ( (spot=strstr(line,"CLIP")) != 0 && frame_number==0 )
  { clip_view_flag = 1;
    retval = sscanf(spot+5,"%lf %lf %lf %lf",init_clip_coeff,init_clip_coeff+1,
        init_clip_coeff+2,init_clip_coeff+3);
    if ( retval != 4 )
    { fprintf(stderr,"\nevmovie error: need 4 clip coefficients in first line of %s after 'CLIP'\n\n",filenames[frame_number]);
      exit(5);
    }
    init_clip_flag = 1;
    // Some fudging on signs to make things work right
    init_clip_coeff[0] *= -1;
    init_clip_coeff[1] *= -1;
    init_clip_coeff[2] *= -1;
  }

  if ( strstr(line,"SMOOTH") && smooth_flag == UNSET )
    smooth_flag = 1;

  if ( strstr(line,"SHOW_EDGES") && edge_show_flag == UNSET )
    edge_show_flag = ON;

  if ( strstr(line,"FACETS_OFF") && frame_number==0 )
    facet_show_flag = 1;

  if ( !view_set_flag && (spot = strstr(line,"VIEW")) != 0 )
  { retval = sscanf(spot+5,"%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",
      init_view[0],init_view[0]+1,init_view[0]+2,init_view[0]+3,
      init_view[1],init_view[1]+1,init_view[1]+2,init_view[1]+3,
      init_view[2],init_view[2]+1,init_view[2]+2,init_view[2]+3,
      init_view[3],init_view[3]+1,init_view[3]+2,init_view[3]+3);
    if ( retval != 16 )
      fprintf(stderr,"\nevmovie error: only %d values for view matrix in %s\n\n",
         retval,filenames[frame_number]);
    else
      view_set_flag = 1;
  }

  spot = line;
  textid = 0;
  while ( (spot = strstr(spot," TEXT ")) != 0 )
  { char *c;
    int n;
    if ( textid >= MAXTEXT )
      break;
    f->text[textid] = (struct text_t *)calloc(1,sizeof(struct text_t));
    sscanf(spot+5,"%lf %lf %lf",&f->text[textid]->x,&f->text[textid]->y,&f->text[textid]->scale);
    while ( *spot && *spot != '"' ) spot++;
    spot++;
    for ( c = f->text[textid]->string, n = 0; *spot && *spot != '"' && n < MAXSTRING; spot++,c++,n++ )
      *c = *spot;
    textid++;
  }

}

void read_file_header_ascii(char *filename, int frame_number)
{ 

#define LINESIZE 2000
  char line[LINESIZE+1];
  struct frame_t *f = frames+frame_number;

  FILE *fd = fopen(filename,"r");
  if ( !fd )
  { perror(filename);
    exit(1);
  }

  fgets(line,LINESIZE,fd); // OFF

  if ( !strstr(line,"OFF") )
  { fprintf(stderr,"%s is not an OFF file; bad first line, no OFF.\n",filename);
    exit(2);
  }

  if ( strstr(line,"BINARY") )
  { // oops, try binary version 
    fclose(fd);
    read_file_header_binary(filename,frame_number);
    return;
  }
 
  read_options(frame_number,line);

  fgets(line,LINESIZE,fd); // element counts
  if ( sscanf(line," %d %d %d",&f->vcount,&f->fcount,&f->ecount) != 3) 
  { fprintf(stderr,"%s is not an OFF file; missing counts on line 2.\n",
       filename);
    exit(2);
  }
 
  fclose(fd);
  
} // end read_file_header_ascii()

void read_file_header_binary(char *filename,int frame_number)
{ char line[LINESIZE];
  struct frame_t *f;
  int should_be_length;

  FILE *fd = fopen(filename,"rb");
  if ( !fd )
  { perror(filename);
    exit(1);
  }

  f = frames+frame_number;

  fgets(line,LINESIZE,fd); // OFF
  if ( !strstr(line,"OFF") )
  { fprintf(stderr,"%s is not an OFF file; bad first line, no OFF.\n",filename);
    exit(2);
  }

  read_options(frame_number,line);

  // get element counts
  fread(&f->vcount,sizeof(int),1,fd);
  fread(&f->fcount,sizeof(int),1,fd);
  fread(&f->ecount,sizeof(int),1,fd);

  f->filelength = _filelength(_fileno(fd));

  // sanity check on file size
  should_be_length = ftell(fd) + f->vcount*3*sizeof(float)
      + f->fcount*(f->opacity_flag?6:5)*sizeof(int) + f->ecount*4*sizeof(int);
  
  if ( should_be_length != f->filelength )
  { if ( f->vcount < 100000 ) // probably got the endian right, so some other problem
    { fprintf(stderr,"\n evmovie error: File %s should be length %d but is %d\n",
            filename,should_be_length,f->filelength);
      fprintf(stderr,"Claims to have %d vertices, %d edges, %d facets.\n",
		 f->vcount,f->ecount,f->fcount);
      if ( should_be_length > f->filelength )
        fprintf(stderr,"Maybe OPACITY option on but facets don't have opacity values?\n");
      else
        fprintf(stderr,"Maybe OPACITY option off but facets have opacity values?\n\n");  
      exit(10);
    }
    // try other-endian
    endian_swap(&(f->vcount),sizeof(int));
    endian_swap(&(f->ecount),sizeof(int));
    endian_swap(&(f->fcount),sizeof(int));
    should_be_length = ftell(fd) + f->vcount*3*sizeof(float)
	    + f->fcount*5*sizeof(int) + f->ecount*4*sizeof(int);
	if ( should_be_length != f->filelength )
    { fprintf(stderr,"File %s should be length %d but is %d\n",
            filename,should_be_length,f->filelength);
      fprintf(stderr,"Claims to have %d vertices, %d edges, %d facets.\n",
		 f->vcount,f->ecount,f->fcount);
      exit(1);
    }
    else f->endian_swap_flag = 1;
  }
  fclose(fd);

} // end read_file_header_binary()

void read_file_ascii(char *filename, int frame_number)
{ char line[LINESIZE];
  struct frame_t *f;
  float *vspot;
  int *fspot;
  int i,j,two,three;

  FILE *fd = fopen(filename,"r");
  if ( !fd )
  { perror(filename);
    exit(1);
  }

  f = frames+frame_number;

  fgets(line,LINESIZE,fd); // OFF
  if ( !strstr(line,"OFF") )
  { fprintf(stderr,"%s is not an OFF file; bad first line, no OFF.\n",filename);
    exit(2);
  }

  if ( strstr(line,"BINARY") )
  { // oops, try binary version 
    fclose(fd);
    read_file_binary(filename,frame_number);
    return;
  }
 
 
  read_options(frame_number,line);


  fgets(line,LINESIZE,fd); // element counts
  if ( sscanf(line," %d %d %d",&f->vcount,&f->fcount,&f->ecount) != 3) 
  { fprintf(stderr,"%s is not an OFF file; missing counts on line 2.\n",filename);
    exit(2);
  }

  // Set memory pointers; f->vertexdata already points to memory frame.
  f->normaldata = f->vertexdata + 3*f->vcount;
  f->facetdata = (int*)(f->normaldata + 3*f->vcount);
  f->colordata = (float*)(f->facetdata + 3*f->fcount);
  f->edgedata = (int*)(f->colordata + 4*f->fcount);

  for ( i = 0, vspot = f->vertexdata ; i < f->vcount ; i++, vspot += 3 )
  { fgets(line,LINESIZE,fd);
    sscanf(line,"%f %f %f",vspot,vspot+1,vspot+2);
    for ( j = 0 ; j < 3 ; j++ )
    { if ( vspot[j] < bbox_min[j] )
        bbox_min[j] = vspot[j];
      if ( vspot[j] > bbox_max[j] )
        bbox_max[j] = vspot[j];
    }
  }

  for ( i = 0, fspot = f->facetdata ; i < f->fcount ; i++, fspot += 3 )
  { int color;
    float opacity;
    fgets(line,LINESIZE,fd);
    sscanf(line,"%d %d %d %d %d %f",&three,fspot,fspot+1,fspot+2,&color,&opacity);
    if ( !f->opacity_flag )
        opacity = 1.0;
    f->colordata[4*i] = rgba[color][0];
    f->colordata[4*i+1] = rgba[color][1];
    f->colordata[4*i+2] = rgba[color][2];
    f->colordata[4*i+3] = opacity;
  }

  for ( i = 0, fspot = f->edgedata ; i < f->ecount ; i++, fspot += 3 )
  { fgets(line,LINESIZE,fd);
    sscanf(line,"%d %d %d %d",&two,fspot,fspot+1,fspot+2);
    if ( fspot[0] >= f->vcount || fspot[1] >= f->vcount || fspot[2] >= f->vcount )
    { fprintf(stderr,"\n\nevmovie FATAL ERROR: A vertex of a facet has index larger than the numbers of vertices.\n\n");
      exit(1);
    }
    if ( fspot[0] < 0 || fspot[1] < 0 || fspot[2] < 0 )
    { fprintf(stderr,"\n\nevmovie FATAL ERROR: A vertex of a facet has a negative index.\n\n");
      exit(1);
    }

  }
  fclose(fd);
  
  color_sort(f);
  calc_normals(f);
  
} // end read_file_ascii()

void read_file_binary(char *filename, int frame_number)
{ char line[LINESIZE];
  struct frame_t *f;
  float *vspot;
  int *fspot;
  int i,j;
  char *spot;

  FILE *fd = fopen(filename,"rb");
  if ( !fd )
  { perror(filename);
    exit(1);
  }

  f = frames+frame_number;
  f->filelength = _filelength(_fileno(fd));

  fgets(line,LINESIZE,fd); // OFF
  if ( !strstr(line,"OFF") )
  { fprintf(stderr,"%s is not an OFF file; bad first line, no OFF.\n",filename);
    exit(2);
  }
 
  read_options(frame_number,line);

  fread(file_buffer,f->filelength,1,fd);

  if ( f->endian_swap_flag )
  { int k;
    for ( k = 0 ; k < f->filelength ; k += sizeof(int))
      endian_swap(file_buffer+k,sizeof(int));
  }
 
  spot = file_buffer;

  f->vcount = *(int*)spot; spot += sizeof(int);
  f->fcount = *(int*)spot; spot += sizeof(int);
  f->ecount = *(int*)spot; spot += sizeof(int);

  
  // Set memory pointers; f->vertexdata already points to memory frame.
  f->normaldata = f->vertexdata + 3*f->vcount;
  f->facetdata = (int*)(f->normaldata + 3*f->vcount);
  f->colordata = (float*)(f->facetdata + 3*f->fcount);
  f->edgedata = (int*)(f->colordata + 4*f->fcount);

  for ( i = 0, vspot = f->vertexdata ; i < f->vcount ; i++, vspot += 3 )
  { vspot[0] = *(float*)spot; spot += sizeof(float);
    vspot[1] = *(float*)spot; spot += sizeof(float);
    vspot[2] = *(float*)spot; spot += sizeof(float);
    for ( j = 0 ; j < 3 ; j++ )
    { if ( vspot[j] < bbox_min[j] )
        bbox_min[j] = vspot[j];
      if ( vspot[j] > bbox_max[j] )
        bbox_max[j] = vspot[j];
    }
  }

  for ( i = 0, fspot = f->facetdata ; i < f->fcount ; i++, fspot += 3 )
  { int color;
    spot += sizeof(int); // skip "3"
    fspot[0] = *(int*)spot; spot += sizeof(int);  // vertex indices
    fspot[1] = *(int*)spot; spot += sizeof(int);
    fspot[2] = *(int*)spot; spot += sizeof(int);
    color = *(int*)spot; spot += sizeof(int);  // color index
    f->colordata[4*i] = rgba[color][0];
    f->colordata[4*i+1] = rgba[color][1];
    f->colordata[4*i+2] = rgba[color][2];
    if ( f->opacity_flag )
    { f->colordata[4*i+3] = *(float*)spot; 
      spot += sizeof(int);
    }
    else
        f->colordata[4*i+3] = 1.0;

    if ( fspot[0] >= f->vcount || fspot[1] >= f->vcount || fspot[2] >= f->vcount )
    { fprintf(stderr,"\n\nevmovie FATAL ERROR: A vertex of a facet has index larger than the numbers of vertices.\n\n");
      exit(1);
    }
    if ( fspot[0] < 0 || fspot[1] < 0 || fspot[2] < 0 )
    { fprintf(stderr,"\n\nevmovie FATAL ERROR: A vertex of a facet has a negative index.\n\n");
      exit(1);
    }
    
  }

 for ( i = 0, fspot = f->edgedata ; i < f->ecount ; i++, fspot += 3 )
  { 
    spot += sizeof(int); // skip "2"
    fspot[0] = *(int*)spot; spot += sizeof(int); // vertex indices
	fspot[1] = *(int*)spot; spot += sizeof(int);
	fspot[2] = *(int*)spot; spot += sizeof(int); // color index
  }

  fclose(fd);
  
  color_sort(f);
  calc_normals(f);
  
} // end read_file_binary()

void resize()
{ double mid[3],size;
  int i,j;
  
  for ( i = 0 ; i < SDIM ; i++ ) 
    mid[i] = (bbox_max[i] + bbox_min[i])/2;
  size = bbox_max[2] - bbox_min[2];
  if ( bbox_max[1] - bbox_min[1] > size ) size = bbox_max[1] - bbox_min[1];
  if ( bbox_max[0] - bbox_min[0] > size ) size = bbox_max[0] - bbox_min[0];
  
  /* transformation matrix will be set up to scale object into
      [-1,1]^3 cube */

  if ( view_set_flag )
  { for ( i = 0 ; i < HOMDIM ; i++ )
      for ( j = 0 ; j < HOMDIM ; j++ )
        view[i][j] = init_view[i][j];
  }
  else
  {
    for ( i = 0 ; i < HOMDIM ; i++ )
      for ( j = 0 ; j < HOMDIM ; j++ )
        view[i][j] = identmat[i][j];
      
    if ( size != 0.0 )
      for ( i = 0 ; i < HOMDIM-1 ; i++ ) 
       { view[i][i] = 2/size;
         view[i][HOMDIM-1] = -mid[i]*2/size;
       }
  }

  if ( !init_clip_flag )
    init_clip_coeff[3] = mid[0];
 
  for ( i = 0 ; i < HOMDIM ; i++ )
    clip_coeff[0][i] = init_clip_coeff[i];
}

void mat_mult(MAT4 a, MAT4 b, MAT4 c, int imax, int jmax, int kmax)
{ MAT4 temp;
  int i,j,k;

  if ( (a == c) || (b == c) )
  { for ( i = 0 ; i < imax ; i++ )
      for ( k = 0 ; k < kmax ; k++ )
      {
        temp[i][k] = 0.0;
        for ( j = 0 ; j < jmax ; j++ )
          temp[i][k] += a[i][j]*b[j][k];
      }
      for ( i = 0 ; i < imax ; i++ )
        for ( k = 0 ; k < kmax ; k++ )
          c[i][k] = temp[i][k];
  }
  else
  { for ( i = 0 ; i < imax ; i++ )
      for ( k = 0 ; k < kmax ; k++ )
      {
        c[i][k] = 0.0;
        for ( j = 0 ; j < jmax ; j++ )
          c[i][k] += a[i][j]*b[j][k];
      }
  }
}

void matvec_mul(MAT4 a, double *b, double *c, int rows, int cols)
{
  int i,j;

  for ( i = 0 ; i < rows ; i++ )
     { c[i] = 0.0;
        for ( j = 0 ; j < cols ; j++ )
          c[i] += a[i][j]*b[j];
     }
}

/******************************************************************
*
* Function: mat_inv()
*
* Purpose: in-place matrix inverse by gauss-jordan 
* returns -1 for singular matrix, 
*          >= 0 for nonsingular, but value is not index, since
*               pivoting is not symmetric. 
*           
*/

#define SWAP(a,b) {double temp = (a); (a) = (b); (b) = temp; }
#define SMALL 10

int mat_inv (
  MAT4 a)     /* 4x4 matrix to invert in place */
{
  int indxc[HOMDIM],indxr[HOMDIM],ipiv[HOMDIM];
  int i,icol=0,irow=0,j,k,l,ll;
  double big,dum,pivinv;
  int retval = 1;  /* default return value is success */
  int n = HOMDIM;

  for ( j = 0 ; j < n ; j++ ) 
    ipiv[j] = -1;
  for ( i = 0 ; i < n ; i++ )
  { /* find pivot */
    big = 0.0;
    for ( j = 0 ; j < n ; j++ )
    { if ( ipiv[j] != 0 )
         for ( k = 0 ; k < n ; k++ )
         { if ( ipiv[k] == -1 )
           { if ( fabs(a[j][k]) >= big )
             { big = fabs(a[j][k]);
               irow = j;
               icol = k;
             }
           }
           else 
             if ( ipiv[k] > 0 ) 
             { return -1;  }
         }
    }
    ++(ipiv[icol]);

    if ( irow != icol )
       for ( l = 0 ; l < n ; l++ ) 
         SWAP(a[irow][l],a[icol][l])
    indxr[i] = irow;
    indxc[i] = icol;
    if ( a[icol][icol] == 0.0 ) 
    { return -1; }
    pivinv = 1/a[icol][icol];
    a[icol][icol] = 1.0;
    for ( l = 0 ; l < n ; l++ ) 
      a[icol][l] *= pivinv;
    for ( ll = 0  ; ll < n ; ll++ )
      if ( ll != icol )
      { dum = a[ll][icol];
        a[ll][icol] = 0.0;
        for ( l = 0 ; l < n ; l++ ) 
          a[ll][l] -= a[icol][l]*dum;
      }
  }
  for ( l = n-1 ; l >= 0 ; l-- )
  { if ( indxr[l] != indxc[l] )
       for ( k = 0 ; k < n ; k++ )
          SWAP(a[k][indxr[l]],a[k][indxc[l]])
  }

  return retval;
} /* end mat_inv() */

double gleft=0,gright=0.05,gtop=0.05,gbottom=0;
double ileft=0.95,iright=1,itop=0.05,ibottom=0;
double dleft=0.05,dright=0.10,dtop=0.05,dbottom=0;
double sleft=0.10,sright=0.95,stop=0.05,sbottom=0;
double bleft=0.10,bright=0.95,btop=0.05,bbottom=0;
double bwidth = 0.05;


void slider()
{
  /* set up window coords */
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(0.0,1.0,0.0,1.0,0.0,1.0);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity(); 
    
  glDisable(GL_LIGHTING);

  glLineWidth(xsize/200);
  // stop-go button
  if ( go_state == GO )
    glColor3f(0,1,0);
  else
    glColor3f(1,0,0);
  glBegin(GL_QUADS);
    glVertex2d(gleft,gtop);
    glVertex2d(gleft,gbottom);
    glVertex2d(gright,gbottom);
    glVertex2d(gright,gtop);
  glEnd();
  glColor3f(0,0,0);
  glBegin(GL_LINE_LOOP);
    glVertex2d(gleft,gtop);
    glVertex2d(gleft,gbottom);
    glVertex2d(gright,gbottom);
    glVertex2d(gright,gtop);
  glEnd();
  glBegin(GL_LINE_STRIP); // direction wedges
    glVertex2d(0.5*gleft + 0.5*gright, 0.7*gtop + 0.3*gbottom);
    glVertex2d(0.2*gleft + 0.8*gright, 0.5*gtop + 0.5*gbottom);
    glVertex2d(0.5*gleft + 0.5*gright, 0.3*gtop + 0.7*gbottom);
  glEnd();
  glBegin(GL_LINE_STRIP); // direction wedges
    glVertex2d(0.8*gleft + 0.2*gright, 0.7*gtop + 0.3*gbottom);
    glVertex2d(0.5*gleft + 0.5*gright, 0.5*gtop + 0.5*gbottom);
    glVertex2d(0.8*gleft + 0.2*gright, 0.3*gtop + 0.7*gbottom);
  glEnd();

  // increment
  glColor3f(1,1,0);
  glBegin(GL_QUADS);
    glVertex2d(ileft,itop);
    glVertex2d(ileft,ibottom);
    glVertex2d(iright,ibottom);
    glVertex2d(iright,itop);
  glEnd();
  glColor3f(0,0,0);
  glBegin(GL_LINE_LOOP);
    glVertex2d(ileft,itop);
    glVertex2d(ileft,ibottom);
    glVertex2d(iright,ibottom);
    glVertex2d(iright,itop);
  glEnd();
  glBegin(GL_LINE_STRIP); // direction wedge
    glVertex2d(0.6*ileft + 0.4*iright, 0.7*itop + 0.3*ibottom);
    glVertex2d(0.3*ileft + 0.7*iright, 0.5*itop + 0.5*ibottom);
    glVertex2d(0.6*ileft + 0.4*iright, 0.3*itop + 0.7*ibottom);
  glEnd();

  // decrement
  glColor3f(1,1,0);
  glBegin(GL_QUADS);
    glVertex2d(dleft,dtop);
    glVertex2d(dleft,dbottom);
    glVertex2d(dright,dbottom);
    glVertex2d(dright,dtop);
  glEnd();
  glColor3f(0,0,0);
  glBegin(GL_LINE_LOOP);
    glVertex2d(dleft,dtop);
    glVertex2d(dleft,dbottom);
    glVertex2d(dright,dbottom);
    glVertex2d(dright,dtop);
  glEnd();
  glBegin(GL_LINE_STRIP); // direction wedge
    glVertex2d(0.4*dleft + 0.6*dright, 0.7*dtop + 0.3*dbottom);
    glVertex2d(0.7*dleft + 0.3*dright, 0.5*dtop + 0.5*dbottom);
    glVertex2d(0.4*dleft + 0.6*dright, 0.3*dtop + 0.7*dbottom);
  glEnd();

  // slider
  glColor3f(0.5,0.5,0.5);
  glBegin(GL_QUADS);
    glVertex2d(sleft,stop);
    glVertex2d(sleft,sbottom);
    glVertex2d(sright,sbottom);
    glVertex2d(sright,stop);
  glEnd();
 

  // slider button
  if ( frame_count <= 1 )
    bleft = sleft;
  else
    bleft = sleft+((double)this_frame/(frame_count-1))*(sright-sleft-bwidth);
  bright = bleft + bwidth;
  glColor3f(1,1,1);
  glBegin(GL_QUADS);
    glVertex2d(bleft,btop);
    glVertex2d(bleft,bbottom);
    glVertex2d(bright,bbottom);
    glVertex2d(bright,btop);
  glEnd();
  glColor3f(0,0,0);
  glBegin(GL_LINE_LOOP);
    glVertex2d(bleft,btop);
    glVertex2d(bleft,bbottom);
    glVertex2d(bright,bbottom);
    glVertex2d(bright,btop);
  glEnd();
  glBegin(GL_LINE_STRIP); // direction wedges
    glVertex2d(0.4*bleft + 0.6*bright, 0.7*btop + 0.3*bbottom);
    glVertex2d(0.1*bleft + 0.9*bright, 0.5*btop + 0.5*bbottom);
    glVertex2d(0.4*bleft + 0.6*bright, 0.3*btop + 0.7*bbottom);
  glEnd();
  glBegin(GL_LINE_STRIP); // direction wedges
    glVertex2d(0.6*bleft + 0.4*bright, 0.7*btop + 0.3*bbottom);
    glVertex2d(0.9*bleft + 0.1*bright, 0.5*btop + 0.5*bbottom);
    glVertex2d(0.6*bleft + 0.4*bright, 0.3*btop + 0.7*bbottom);
  glEnd();
 
  glEnable(GL_LIGHTING);

} // end slider()
  
/***********************************************************
*
* function: mouse_func()
*
* purpose: Called on mouse button events, records position.
*/

void mouse_func (
int button, int state, int x, int y)
{ static int down_in_go_button;
  switch ( button )
  { case GLUT_LEFT_BUTTON:
      mouse_left_state = state;
      if ( state == GLUT_UP )
        mouse_mode = last_mouse_mode;
      if ( y > (1-dtop)*ysize )
      { double xspot = (double)x/xsize;
        if ( xspot < gright )
        { if ( state == GLUT_UP )
		  { if ( down_in_go_button )
		    { go_state = (go_state == GO) ? STOP : GO;
		      mouse_mode = last_mouse_mode;
		      glutPostRedisplay();
		      down_in_go_button = 0;
		    }
		  }
		  else 
		  { down_in_go_button = 1;
            if ( loop_mode == UNSET )
              loop_mode = NOLOOP;
            if ( loop_mode == NOLOOP && this_frame == frame_count-1 )
              this_frame = 0;
		    mouse_mode = MM_GO_BUTTON;
		  }
        }
        else if ( xspot < dright )
        { if ( state == GLUT_UP )
          { if ( this_frame > 0 )
            { this_frame--;
              glutPostRedisplay();
            } 
          }
          else 
              mouse_mode = MM_LEFT_ARROW;
        }
        else if ( xspot > ileft )
        { if ( state == GLUT_UP )
          { if ( this_frame < frame_count-1 )
            { this_frame++;
              glutPostRedisplay(); 
            } 
          } 
          else
             mouse_mode = MM_RIGHT_ARROW;
        }
        else if ( (xspot > bleft) && (xspot < bright) )
        { 
          if ( state == GLUT_DOWN ) 
          {
            mouse_mode = MM_SLIDER;
            this_frame = (int)((xspot-sleft)/(sright-sleft)*frame_count);
            glutPostRedisplay(); 
 
            oldx = x;          
          }
          else
          { mouse_mode = last_mouse_mode;

          }
        }
        else // go to slider spot
        { this_frame = (int)((xspot-sleft)/(sright-sleft)*frame_count);
          
          if ( state == GLUT_UP )
            mouse_mode = last_mouse_mode;
          else
            mouse_mode = MM_SLIDER;
          glutPostRedisplay(); 
        }
      }
      else
      switch ( state )
      { case GLUT_DOWN:  /* start tracking */
           oldx = x; oldy = y;
           mouse_left_state = GLUT_DOWN;
           glutIdleFunc(idle_func);
           break;
        case GLUT_UP:  /* stop tracking */
           basex = newx;
           basey = newy;
           mouse_left_state = GLUT_UP;
           glutPostRedisplay();
           glutIdleFunc(idle_func);
           break;
      }
      break;

    case GLUT_RIGHT_BUTTON:
      switch ( state )
      { case GLUT_DOWN: 
          glutPostRedisplay();  /* get image back */  
          break;
      }
      break;
  }
}

/********************************************************************
*
*  function: fix_ctm()
*
*    Rotates coordinate transformation matrix according to how mouse 
*    dragged.
*/

void fix_ctm(viewmat,dx,dy)
MAT4 viewmat; /* matrix to modify */
double dx,dy;  /* pixels mouse dragged */
{
  MAT4 rot;
  double alpha;  /* angle around axis */
  double theta;    /* tilt of rotation axis */
  int i,j;
  
  for ( i = 0 ; i < HOMDIM ; i++ )
  { for ( j = 0 ; j < HOMDIM ; j++ )
        rot[i][j] = 0.0;
    rot[i][i] = 1.0;
  }
  alpha = sqrt(dx*dx + dy*dy)/300;  /* one radian per 300 pixels */
  if ( dx == 0.0 ) 
  { if ( dy > 0.0 ) theta = M_PI/2;
    else if ( dy < 0.0 ) theta = -M_PI/2;
    else goto ctm_exit;  /* no change */
  }
  else 
  { theta = atan(dy/dx);
    if ( dx < 0.0 ) alpha = - alpha;
  }
  
  if ( SDIM == 2 )
  { 
    /* tilt axis */
    rot[2][2] = 1.0;
    rot[0][0] = rot[1][1] = cos(theta);
    rot[0][1] = sin(theta);
    rot[1][0] = -sin(theta);
    mat_mult(rot,viewmat,viewmat,HOMDIM,HOMDIM,HOMDIM);
  
    /* rotate */
    rot[2][2] = rot[0][0] = cos(alpha);
    rot[2][0] = -sin(alpha);
    rot[0][2] = sin(alpha);
    rot[1][1] = 1.0;
    rot[0][1] = rot[1][0] = 0.0;
    mat_mult(rot,viewmat,viewmat,HOMDIM,HOMDIM,HOMDIM);
  
    /* untilt axis */
    rot[2][2] = 1.0;
    rot[0][0] = rot[1][1] = cos(theta);
    rot[0][1] = -sin(theta);
    rot[1][0] = sin(theta);
    rot[2][0] = rot[0][2] = 0.0;
  }
  else
  {
    /* tilt axis */
    rot[0][0] = 1.0;
    rot[1][1] = rot[2][2] = cos(theta);
    rot[1][2] = sin(theta);
    rot[2][1] = -sin(theta);
    mat_mult(rot,viewmat,viewmat,HOMDIM,HOMDIM,HOMDIM);
  
    /* rotate */
    rot[0][0] = rot[1][1] = cos(alpha);
    rot[0][1] = -sin(alpha);
    rot[1][0] = sin(alpha);
    rot[2][2] = 1.0;
    rot[1][2] = rot[2][1] = 0.0;
    mat_mult(rot,viewmat,viewmat,HOMDIM,HOMDIM,HOMDIM);
  
    /* untilt axis */
    rot[0][0] = 1.0;
    rot[1][1] = rot[2][2] = cos(theta);
    rot[1][2] = -sin(theta);
    rot[2][1] = sin(theta);
    rot[0][1] = rot[1][0] = 0.0;
  }
  mat_mult(rot,viewmat,viewmat,HOMDIM,HOMDIM,HOMDIM);
  
ctm_exit: ;
} /* end fix_ctm() */

/*******************************************************************
*
* function: mouse_loc_func()
*
* purpose: Called as mouse moves with left button down, this
*          moves surface according to current mouse_mode.
*/

void mouse_loc_func (
int x, int y)
{ 
  int i,j;

  newx = x;
  newy = y; 
  if ( mouse_left_state == GLUT_DOWN )
  { switch ( mouse_mode )
    { case MM_GO_BUTTON: return; // do nothing
      case MM_LEFT_ARROW: return; // do nothing
      case MM_RIGHT_ARROW: return; // do nothing

      case MM_SLICE: // clipping, actually
         if ( clip_view_flag )
         { clip_coeff[0][SDIM] += (newx-oldx)*xscale/5;
           
           break;
         }
         else
         /* reset back to rotate by default and fall through */
           last_mouse_mode = mouse_mode = MM_ROTATE;
 
      case MM_ROTATE:    
       fix_ctm(view,(double)( newx - oldx), -(double)(newy - oldy));
       break;

      case MM_SCALE:
        for(i = 0 ; i < HOMDIM-1; i++ )
            for ( j = 0 ; j < HOMDIM ; j++ )
                view[i][j] *= 1.0 + 0.002*(newx-oldx);
        clip_coeff[0][SDIM] *= 1.0 + 0.004*(newx-oldx);
        // 0.004 instead of 0.002 since rhs of clip plane
        // is the square distance.
        break;

      case MM_TRANSLATE:
          view[1][HOMDIM-1] -= (newx-oldx)*xscale;
          view[2][HOMDIM-1] += (newy-oldy)*yscale;   
          break;

      case MM_SPIN: /* about z axis */
        { MAT4 rot;
          double dth;
          double dang; /* fourth dimension */
          for ( i = 0 ; i < HOMDIM ; i++ )
          { for ( j = 0 ; j < HOMDIM ; j++ )
              rot[i][j] = 0.0;
            rot[i][i] = 1.0;
          }
          dth = (newx - oldx)/300.0*M_PI;
          dang = (newy - oldy)/300.0*M_PI;
          if ( SDIM == 2 )
          { rot[0][0] = rot[1][1] = cos(dth);
            rot[0][1] = -(rot[1][0] = sin(dth));
          } else
          { rot[1][1] = rot[2][2] = cos(dth);
            rot[1][2] = -(rot[2][1] = sin(dth));
          } 
          mat_mult(rot,view,view,HOMDIM,HOMDIM,HOMDIM);
        }
        break;

      case MM_SLICE_SPIN:
        {
          double dth;
          double dang; 
          double temp1[3];
          double temp2[3];

          dang = (newx - oldx)/300.0*M_PI;
          dth = -(newy - oldy)/300.0*M_PI;

          matvec_mul(from_focus,clip_coeff[0],temp1,3,3);
          temp2[0] = cos(dth)*temp1[0] - sin(dth)*temp1[2];
          temp2[1] = temp1[1];
          temp2[2] = sin(dth)*temp1[0] + cos(dth)*temp1[2];
          temp1[2] = temp2[2];
          temp1[0] = cos(dang)*temp2[0] - sin(dang)*temp2[1];
          temp1[1] = sin(dang)*temp2[0] + cos(dang)*temp2[1];
          matvec_mul(to_focus,temp1,clip_coeff[0],3,3);
 
          break;
        }
        
        case MM_SLIDER:
        { double xspot = (double)x/xsize;
          this_frame = (int)((xspot-sleft)/(sright-sleft)*frame_count);
          break;
        }

      }

      glutPostRedisplay();
    }

  oldx = newx; oldy = newy;
}

/************************************************************************
*
* function: reshape_func()
*
* purpose: Handle window resize messages, and set up screen coordinates
*          to compensate for window aspect ratio.
*/

void reshape_func(
int x, int y)
{ 
  if ( window_aspect_ratio != 0.0 )
  { /* munge x,y */
    double area = x*y;
    
    /* see if moving just one side of frame */
    if ( abs(xsize - x) < abs(ysize - y) )
    { /* moving y */
      x = (int)(y/fabs(window_aspect_ratio));
    }
    else 
    { /* moving x */
      y = (int)(x*fabs(window_aspect_ratio));
    }
    window_aspect = window_aspect_ratio;
    aspect_flag = 0;
    glutReshapeWindow(x,y);
  }
  
  xsize = x; ysize = y;
  aspect = (double)y/x;
  glViewport(0,0,x,y);
  if ( aspect > 1 ) 
  { xscale = 2.8/aspect/x; yscale = 2.8/y; 
    imagescale = yscale*100;
  }
  else 
  { xscale = 2.8/x; yscale = 2.8*aspect/y; 
    imagescale = xscale*100;
  }
   
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  
  if ( aspect >= 1.0 )
  { glOrtho(scrx[0],scrx[2],aspect*scry[0],
             aspect*scry[2],-50.0,50.0);      
  }
  else
  { glOrtho(scrx[0]/aspect,scrx[2]/aspect,scry[0],
          scry[2],-50.0,50.0);
  }
}  // end reshape_func()

void idle_func(void)
{
  if ( go_state == GO )
  { if ( loop_mode == NOLOOP )
    { if ( this_frame < frame_count-1 )
        this_frame++;
      else 
        go_state = STOP; 
    }
    else if ( loop_mode == LOOP )
    { if ( this_frame < frame_count-1 )
        this_frame++;
      else 
        this_frame = 0;
    }
    else if ( loop_mode == CYCLE )
    { if ( go_direction == FORWARD )
      { if ( this_frame < frame_count-1 )
          this_frame++;
        else 
          go_direction = BACKWARD;
          
      }
      else
      { if ( this_frame > 0 )
          this_frame--;
        else
          go_direction = FORWARD;
      }
    }
    glutPostRedisplay();
  }
}

/*****************************************************************************
*
* function glut_text_display()
*
* purpose: display text lines in graphics window.
*/

void glut_text_display()
{ char *c;
  int i;
  struct frame_t *f = frames + this_frame;
  void *font;
  double pixheight;

  /* set up window coords */
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(0.0,1.0,0.0,1.0,0.0,1.0);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();

  
  for ( i = 0 ; i < MAXTEXT ; i++ )
  { double yspot;
    int lcount = 0;
    if ( f->text[i] == NULL )
      continue;
    yspot = f->text[i]->y;
    glColor3f(0.0,0.0,0.0); /* black */
    pixheight = f->text[i]->scale*ysize;
    if ( pixheight < 8 || pixheight > 25 )
    { // do stroke text
       glLoadIdentity();
       glTranslatef(f->text[i]->x,f->text[i]->y+gtop,0.0);
       glScalef(.0005*aspect*f->text[i]->scale/.10,.0005*f->text[i]->scale/.08,1.);  // stroke characters are big, 120 units high
       glLineWidth(0.10*f->text[i]->scale*ysize);  // line width in pixels
       for ( c = f->text[i]->string ; *c ; c++ )
       {
         if ( *c == '\n' )
         { lcount++;
           glLoadIdentity();
           glTranslatef(f->text[i]->x,f->text[i]->y+gtop - lcount*1.3*f->text[i]->scale,0.0);
           glScalef(.0005*aspect*f->text[i]->scale/.10,.0005*f->text[i]->scale/.08,1.);  // stroke characters are big, 120 units high
         }
         glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN,*c);
       }
       glLineWidth(linewidth);
    }
    else // pick close bitmap font
    { if ( pixheight < 11 )
        font = GLUT_BITMAP_HELVETICA_10;
      else if ( pixheight < 13 )
        font = GLUT_BITMAP_HELVETICA_12;
      else if ( pixheight < 20 )
        font = GLUT_BITMAP_HELVETICA_18;
      else
        font = GLUT_BITMAP_TIMES_ROMAN_24;
      glLoadIdentity();
      glRasterPos2d(f->text[i]->x,f->text[i]->y+gtop);
      for ( c = f->text[i]->string ; *c ; c++ )
      {  if ( *c == '\n' )
        { yspot -= 18.0/ysize;
          glRasterPos2d(f->text[i]->x,yspot);
        }
        glutBitmapCharacter(font,*c);
      }
    }
  }  // end for
 
  glPopMatrix();  
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
 
} // end glut_text_display()

struct depth_s { int index; double depth; } *depth_sort;

int depth_comp(const void *a, const void *b)
{ if ( ((struct depth_s*)a)->depth < ((struct depth_s*)b)->depth ) return -1;
  if ( ((struct depth_s*)a)->depth > ((struct depth_s*)b)->depth ) return  1;
  return 0;
}

void draw_screen(void)
{ int i,j,c,spot,nspot;
  float viewf[4][4];
  struct frame_t *f;
  char title[200];
	  
  if ( this_frame < 0 ) this_frame = 0;
  else if ( this_frame >= frame_count ) this_frame = frame_count - 1;
  f = frames + this_frame;

  load_frame(this_frame);
  sprintf(title,"evmovie, frame %04d - right click for menu",this_frame+1);
  glutSetWindowTitle(title);
  
  glVertexPointer(3,GL_FLOAT,0,frames[this_frame].vertexdata);
  glNormalPointer(GL_FLOAT,0,frames[this_frame].normaldata);
  glColorPointer(4,GL_FLOAT,0,frames[this_frame].colordata);
 
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  
  glDisable(GL_BLEND);
  slider();
  
  glMatrixMode(GL_PROJECTION);
  glLoadMatrixd(flip[0]);
  glOrtho(-4.0,4.0,-1.4,1.4,-1.4,1.4);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();

  for ( i = 0 ; i < 4 ; i++ )
    for ( j = 0 ; j < 4 ; j++ )
      viewf[i][j] = view[j][i];
  glLoadMatrixf(viewf[0]); 
  
  if ( clip_view_flag )
  { 
     glEnable(GL_CLIP_PLANE0);
     glClipPlane(GL_CLIP_PLANE0,clip_coeff[0]);       
  }

  glColor3f(1.0,1.0,1.0);
  glLineWidth(linewidth);

  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);
  glEnableClientState(GL_COLOR_ARRAY);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  if ( translucency_flag || (f->opacity_flag && opacity_enable) ) 
  { glEnable(GL_BLEND);
    glDisable(GL_DEPTH_TEST);  // otherwise very patchy
    // set up depth sort, just using first vertex of facet
    for ( i = 0 ; i < f->fcount ; i++ )
    { 
      depth_sort[i].index = i;
      depth_sort[i].depth = 0.0;
      for ( j = 0 ; j < 3 ; j++ )  // centroid of facet
        depth_sort[i].depth += (view[0][0]*f->vertexdata[3*f->facetdata[3*i+j]]
                          + view[0][1]*f->vertexdata[3*f->facetdata[3*i+j]+1]
                          + view[0][2]*f->vertexdata[3*f->facetdata[3*i+j]+2]);
    }
    qsort(depth_sort,f->fcount,sizeof(struct depth_s),depth_comp);

  }
  else 
  { glDisable(GL_BLEND);
    glEnable(GL_DEPTH_TEST);
  }

  { float *colorp;
    int opaflag =  translucency_flag || (f->opacity_flag && opacity_enable);
    float old_op; // save native opacity in case of translucency option

      for ( i = 0, spot=0; i < f->fcount ; i++ )
	  { float *t,*h,*hh,nx[3];
        if ( opaflag )
        { spot = 3*depth_sort[i].index;
          t = f->vertexdata + 3*f->facetdata[spot];
          h = f->vertexdata + 3*f->facetdata[spot+1];
          hh = f->vertexdata + 3*f->facetdata[spot+2];
          colorp = f->colordata + 4*depth_sort[i].index;
          if ( translucency_flag )
          { old_op = colorp[3];
            colorp[3] = 0.5;
          }
        }
        else
        { t = f->vertexdata + 3*f->facetdata[spot++];
          h = f->vertexdata + 3*f->facetdata[spot++];
          hh = f->vertexdata + 3*f->facetdata[spot++];
          colorp = f->colordata + 4*i;
        }
		if ( facet_show_flag == ON )
		{ if (smooth_flag == ON)
          { glColor4fv(colorp);
            nspot = opaflag ? 3*depth_sort[i].index : 3*i;
            glBegin(GL_TRIANGLES);
		     glNormal3fv(&f->normaldata[3*f->facetdata[nspot++]]);glVertex3fv(t);
		     glNormal3fv(&f->normaldata[3*f->facetdata[nspot++]]);glVertex3fv(h);
		     glNormal3fv(&f->normaldata[3*f->facetdata[nspot]]);glVertex3fv(hh);
            glEnd();
          }
          else 
          { double mag;
            nx[0] = (h[1]-t[1])*(hh[2]-t[2]) - (h[2]-t[2])*(hh[1]-t[1]);
            nx[1] = (h[2]-t[2])*(hh[0]-t[0]) - (h[0]-t[0])*(hh[2]-t[2]);
            nx[2] = (h[0]-t[0])*(hh[1]-t[1]) - (h[1]-t[1])*(hh[0]-t[0]);
            mag = sqrt(nx[0]*nx[0] + nx[1]*nx[1] + nx[2]*nx[2]);
            nx[0] /= mag; nx[1] /= mag; nx[2] /= mag;
		    glNormal3fv(nx);
            glColor4fv(colorp);
            glBegin(GL_TRIANGLES);
		    glVertex3fv(t);
		    glVertex3fv(h);
		    glVertex3fv(hh);
            glEnd();
          }
		}
        if ( translucency_flag )
          colorp[3] = old_op;

		if ( edge_show_flag == ON )
		{ glColor4f(0.0,0.0,0.0,0.5);
          //glPushMatrix();   // somehow freezes everything but one facet
          glMatrixMode(GL_PROJECTION);
          glTranslated(edge_bias,0.0,0.0);
		  glBegin(GL_LINE_LOOP);
		    glVertex3fv(t);
			glVertex3fv(h);
			glVertex3fv(hh);
		  glEnd();
          glTranslated(-edge_bias,0.0,0.0);
          //glPopMatrix();
		}
	  } // end facet loop     

  } // end flat shading
   
  // permanent edges
  glDisableClientState(GL_COLOR_ARRAY);
  glDisable(GL_LIGHTING);
//  glEnable(GL_DEPTH_TEST);
  glDisable(GL_BLEND);
  glMatrixMode(GL_PROJECTION);
  glTranslated(2*edge_bias,0.0,0.0);  // so get in front of grid edges
  for ( c = 0, spot = 0 ; c < 16 ; c++ )
  { if ( f->ec_counts[c] == 0 ) continue;
    glColor3fv(rgba[c]);
    glDrawElements(GL_LINES,2*f->ec_counts[c],GL_UNSIGNED_INT,
       f->edgedata+spot);
    spot += 2*f->ec_counts[c];
  }
  glTranslated(-edge_bias,0.0,0.0);
  glEnable(GL_LIGHTING);

  glDisable(GL_CLIP_PLANE0);

   // bounding box
  if ( bounding_box_flag==ON )
  { glColor3f(0.0,0.0,0.0);
    glBegin(GL_LINE_LOOP);
      glVertex3d(bbox_min[0],bbox_min[1],bbox_min[2]);
      glVertex3d(bbox_max[0],bbox_min[1],bbox_min[2]);
      glVertex3d(bbox_max[0],bbox_max[1],bbox_min[2]);
      glVertex3d(bbox_min[0],bbox_max[1],bbox_min[2]);
    glEnd();
    glBegin(GL_LINE_LOOP);
      glVertex3d(bbox_min[0],bbox_min[1],bbox_max[2]);
      glVertex3d(bbox_max[0],bbox_min[1],bbox_max[2]);
      glVertex3d(bbox_max[0],bbox_max[1],bbox_max[2]);
      glVertex3d(bbox_min[0],bbox_max[1],bbox_max[2]);
    glEnd();
    glBegin(GL_LINES);
      glVertex3d(bbox_min[0],bbox_min[1],bbox_min[2]);
      glVertex3d(bbox_min[0],bbox_min[1],bbox_max[2]);

      glVertex3d(bbox_max[0],bbox_min[1],bbox_min[2]);
      glVertex3d(bbox_max[0],bbox_min[1],bbox_max[2]);

      glVertex3d(bbox_max[0],bbox_max[1],bbox_min[2]);
      glVertex3d(bbox_max[0],bbox_max[1],bbox_max[2]);

      glVertex3d(bbox_min[0],bbox_max[1],bbox_min[2]);
      glVertex3d(bbox_min[0],bbox_max[1],bbox_max[2]);
    glEnd();
  }

  glut_text_display();

  // delay if necessary
  now = clock();
  if ( now < then + ticks_per_frame )
    Sleep(((then + ticks_per_frame) - now)*msec_per_tick);
  then = clock();

  glFlush();
  glutSwapBuffers();

}

/***********************************************************************
*
* Function: key_func()
*
* purpose: handle ASCII keystrokes in graphics window.
*/

void key_func(
unsigned char key,
int x, int y)
{ 

  switch ( key )
  { 
    case 'n': /* slower movie */
              ticks_per_frame *= 2;
              break;

    case 'm': /* faster movie */
              if ( ticks_per_frame > 1)  
                  ticks_per_frame /= 2;
              break;

    case 'F': /* forward movie */
              go_state = GO;
              loop_mode = NOLOOP;
              go_direction = FORWARD;
              this_frame = 0;
              break;

    case 'O': /* keep looping back to the start */
              go_state = GO;
              loop_mode = LOOP;
              go_direction = FORWARD;
              break;

    case 'C': /* cycle back and forth */
              go_state = GO;
              loop_mode = CYCLE;
              break;

    case 'S': /* stop movie */
              go_state = STOP;
              loop_mode = NOLOOP;
              break;

    case '+': if ( this_frame < frame_count - 1 )
                this_frame++;
              break;

    case '-': if ( this_frame > 0 )
                this_frame--;
              break;

    case 'o': opacity_enable = !opacity_enable; break;

    case 'T': translucency_flag = !translucency_flag; break;

    case 'l':
              glEnable(GL_CLIP_PLANE0);
              clip_view_flag = 1;  /* turn clip_view on */ 
              last_mouse_mode = mouse_mode = MM_SLICE;             
              break;
              
    case 'L': glDisable(GL_CLIP_PLANE0);
              clip_view_flag = 0;
              last_mouse_mode = mouse_mode = MM_ROTATE;
              break;

    case 'G': switch ( background_color )
              { case LIGHTBLUE:  background_color = WHITE; break;
                  case WHITE:  background_color = LIGHTGRAY; break;
                  case LIGHTGRAY: background_color = LIGHTBLUE; break;
              }

              glClearColor(rgba[background_color][0],
                  rgba[background_color][1],
                  rgba[background_color][2],
                  rgba[background_color][3]
              );
              break;

    case 'b': bounding_box_flag = bounding_box_flag==ON ? OFF : ON; break;
         
    case 'r':  last_mouse_mode = mouse_mode = MM_ROTATE; break;
    case 't':  last_mouse_mode = mouse_mode = MM_TRANSLATE; break;
    case 'z':  last_mouse_mode = mouse_mode = MM_SCALE; break;
    case 'c':  last_mouse_mode = mouse_mode = MM_SPIN; break;
    case 'k':  clip_view_flag = 1;  /* turn clip_view on */ 
               last_mouse_mode = mouse_mode = MM_SLICE_SPIN;
               break;
 
    case 'e': edge_show_flag = edge_show_flag==ON ? OFF : ON; break;
    case 'f': facet_show_flag = facet_show_flag==ON ? OFF : ON; break;
    case 'g': smooth_flag = smooth_flag==ON ? OFF : ON; break;
    
    case 'w':
        if ( linewidth > 0.6 )
        { linewidth -= 0.5;  
          glLineWidth(linewidth);
        } 
        break;
  
    case 'W':
        if ( linewidth < 9.9 ) 
        { linewidth += 0.5; 
          glLineWidth(linewidth);
        }
        break;

    case 'R':
        clip_coeff[0][0] = -1.0;
        clip_coeff[0][1] = clip_coeff[0][2] = clip_coeff[0][3] = 0.0;
        clip_view_flag = 0;
        view_set_flag = 0;
        resize(); 
		mouse_mode = MM_ROTATE;
        last_mouse_mode = MM_ROTATE; // remember during sliding
        break;
        
    case 'x': case 'q': exit(0);
 
  }
  glutPostRedisplay();  /* generate redraw message */
} /* end of key_func() */

 void mainmenu_func (int choice)
 { 
   key_func(choice,0,0); 
 } 

void myMenuInit()
 { mainmenu = glutCreateMenu(mainmenu_func);
   glutSetMenu(mainmenu);

   glutAddMenuEntry("Forward movie (F)",'F');
   glutAddMenuEntry("Loop movie (O)",'O');
   glutAddMenuEntry("Cycle back and forth (C)",'C');
   glutAddMenuEntry("Halt movie (S)",'S');
   glutAddMenuEntry("Faster movie (m)",'m');
   glutAddMenuEntry("Slower movie (n)",'n');
   glutAddMenuEntry("Halt movie (S)",'S');
   glutAddMenuEntry("Rotate mode (r)",'r');
   glutAddMenuEntry("Translate mode (t)",'t');
   glutAddMenuEntry("Spin mode (c)",'c');
   glutAddMenuEntry("Scale mode (z)",'z');
   glutAddMenuEntry("Clip plane shift  (l)",'l');
   glutAddMenuEntry("Clip plane rotate (k)",'k');
   glutAddMenuEntry("Disable clip mode (L)",'L');
   glutAddMenuEntry("Next frame (+)",'+');
   glutAddMenuEntry("Previous frame (-)",'-');
   glutAddMenuEntry("Bounding box toggle(b)",'b');
   glutAddMenuEntry("Smooth shading toggle (g)",'g');
   glutAddMenuEntry("Translucency toggle (T)",'T');
   glutAddMenuEntry("Opacity toggle (o)",'o');
   glutAddMenuEntry("Cycle background color (G)",'G');
   glutAddMenuEntry("Edges thicker (W)",'W');
   glutAddMenuEntry("Edges thinner (w)",'w');
   glutAddMenuEntry("Toggle edges (e)",'e');
   glutAddMenuEntry("Toggle faces (f)",'f');
   glutAddMenuEntry("Reset graphics (R)",'R');
   glutAddMenuEntry("Close program (q,x)",'q');
   glutSetMenu(mainmenu);
   glutAttachMenu(GLUT_RIGHT_BUTTON);
 }

void load_frame(int frame_number)
{ struct frame_t *f = frames + frame_number;

  f->timestamp = global_timestamp++;
  if ( f->loaded_flag ) return;

  // Get memory.
  if ( free_memory_count )
  { // Take off free frame list
	f->vertexdata = free_memory_frames[--free_memory_count];
  }
  else 
  { // preempt memory of frame, least-recently-used
    // Search all frames for LRU
    int i,oldframe;
    int old_endian_swap_flag;
	struct frame_t *ff;
	int oldstamp = global_timestamp;
	for ( i = 0 ; i < frame_count ; i++ )
	{ ff = frames + i;
	  if ( ff->loaded_flag && (ff->timestamp < oldstamp) )
	  { oldstamp = ff->timestamp;
	    oldframe = i;
	  }
	}
	// Transfer old frame memory to new frame
	ff = frames + oldframe;
	old_endian_swap_flag = f->endian_swap_flag;
	memset(f,0,sizeof(struct frame_t));
	f->vertexdata = ff->vertexdata;
	memset(f->vertexdata,0,maxmem);
    f->timestamp = global_timestamp++;
    f->endian_swap_flag = old_endian_swap_flag;
	ff->loaded_flag = 0;
	ff->vertexdata = NULL;
	ff->edgedata = NULL;
	ff->facetdata = NULL;
    ff->colordata = NULL;
	ff->normaldata = NULL;
  }

  if ( binary_flag )
      read_file_binary(filenames[frame_number],frame_number);
    else
      read_file_ascii(filenames[frame_number],frame_number);
  
  f->loaded_flag = 1;
}

void usage()
{
  fprintf(stderr,"Usage:  evmovie filelist\n");
  fprintf(stderr,"   where filelist is a list of off images or binary off images.\n");
}

int namecomp(const char **a, const char **b)
{ int retval = strcmp(*a,*b);
  return retval;
}

main(int argc, char **argv)
{ int i;
  int memory_print_flag = 0;
  int retval;
  char findname[1000];
  WCHAR widefindname[1000];

  fprintf(stderr,"evmovie version 1.2.1\n");

  if ( sizeof(int) != sizeof(float) )
  { fprintf(stderr,"This machine has sizeof(int) = %d, sizeof(float) = %d.\n",sizeof(int),sizeof(float));
    fprintf(stderr,"But evmovie assumes they are the same size, so fatal error.\n");
    exit(2);
  }
  
  glutInit(&argc,argv);

  // Own option processing
  while ( argc > 1 && argv[1][0] == '-' )
  { switch ( argv[1][1] )
    { case 'b': binary_flag = 1; break;
      case 'm': memory_print_flag = 1; break;
      
      default: fprintf(stderr,"Unknown option.\n"); usage();exit(100);
    }
    argv++; argc--;
  }

  // Check again for image files.
  if ( argc < 2 )
  { fprintf(stderr,"No image files on command line.\n");
    usage();
    exit(1);
  }

  
  // See if file argument is a folder
  MultiByteToWideChar(CP_ACP,0,argv[1],-1,widefindname,1000);
  retval = GetFileAttributes(widefindname);
  if ( retval == INVALID_FILE_ATTRIBUTES )
  { fprintf(stderr,"evmovie: Cannot find '%s'\n",argv[1]);
    exit(3);
  }
  if ( retval & FILE_ATTRIBUTE_DIRECTORY )
  { // have directory as argument
    HANDLE fh;
    WIN32_FIND_DATA filedata;
    char foundname[256];

    strncpy(findname,argv[1],980);
    strcat(findname,"/*.off*");
    MultiByteToWideChar(CP_ACP,0,findname,1000,widefindname,1000);
    frame_count  = 0;
    fh = FindFirstFile(widefindname,&filedata);
    if ( fh == INVALID_HANDLE_VALUE )
    { fprintf(stderr,"\nevmovie: Cannot find any image files (*.off*) in '%s'\n\n",
         argv[1]);
      exit(2);
    }
    do
    { int namesize;
#ifdef UNICODE
      namesize = WideCharToMultiByte(CP_ACP,0,filedata.cFileName,-1,filenames[frame_count],0,NULL,NULL);
      filenames[frame_count] = (char*)calloc(1,strlen(argv[1])+namesize+10);
      WideCharToMultiByte(CP_ACP,0,filedata.cFileName,-1,foundname,sizeof(foundname),NULL,NULL);
#else
      filenames[frame_count] = calloc(1,strlen(argv[1])+strlen(filedata.cFileName)+10);
      strcpy(foundname,filedata.cFileName);
#endif
      strcpy(filenames[frame_count],argv[1]);
      strcat(filenames[frame_count],"/");
      strcat(filenames[frame_count],foundname);
      frame_count++;
    }
    while ( FindNextFile(fh,&filedata) );

  }
  else
  { // copy from argv
    frame_count = 0;
    while ( argv[1] )
    { filenames[frame_count++] = argv[1];
      argv++;
    }
  }
    
  // and alphabetize names
  qsort(filenames,frame_count,sizeof(char*),namecomp);
  
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
  glutInitWindowPosition(20,20);
  glutInitWindowSize(xsize,ysize);
  glutCreateWindow("evmovie, frame 0001 - right click for menu.");

  myMenuInit();
  glutMouseFunc(mouse_func);
  glutMotionFunc(mouse_loc_func);
  glutKeyboardFunc(key_func);
  glutReshapeFunc(reshape_func);
  glutIdleFunc(idle_func);
  glutShowWindow();  /* start on top */

  glDepthFunc(GL_LEQUAL);
  glEnable(GL_DEPTH_TEST);


  /* clear screen and zbuffer */
  glClearColor(rgba[background_color][0],
          rgba[background_color][1],
          rgba[background_color][2],
          rgba[background_color][3]
          );


  glEnable(GL_NORMALIZE);
  glEnable(GL_COLOR_MATERIAL);
  glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, mat_white);

  glEnable(GL_LIGHTING);
  glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
  glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambient);
  glLightfv(GL_LIGHT1, GL_POSITION, light1_position);
  glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_none);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);

  /* screen corners */
  scrx[0] = scrx[1] = -1.4;
  scrx[2] = scrx[3] =  1.4;
  scry[0] = scry[3] =  1.4;
  scry[1] = scry[2] = -1.4;

  // See how much memory we can use.
  #if defined(WIN32) || defined(_WIN32)
  {
    MEMORYSTATUSEX statex;
    statex.dwLength = sizeof (statex);
    GlobalMemoryStatusEx (&statex);
    if ( memory_print_flag )
    {
      printf ("There are %d total Mbytes of physical memory.\n",
         (int)(statex.ullTotalPhys/1024/1024));
      printf ("There are %d total Mbytes of avialable physical memory.\n",
         (int)(statex.ullAvailPhys/1024/1024));
      printf ("There are %d total Mbytes of available virtual memory.\n",
         (int)(statex.ullAvailVirtual/1024/1024));
    }
    if ( statex.ullAvailVirtual < statex.ullAvailPhys )
      available_bytes = statex.ullAvailVirtual;
    else
      available_bytes = statex.ullAvailPhys;
 }
 #else
 #ifdef _SC_PHYS_PAGES
  // Use POSIX sysconf() to find memory size.
  { long int pages = sysconf(_SC_PHYS_PAGES);
    long int pagesize = sysconf(_SC_PAGE_SIZE);
    available_bytes = (double)pages*(double)pagesize;
    if ( memory_print_flag )
      printf("There are %ld pages of size %ld, total %g bytes, of memory.\n",
         pages,pagesize,available_bytes);
  }
 #else
  available_bytes = 1 << 30;  // default to 1 GB for now
 #endif
 #endif

  // read file header data
  
  frames = (struct frame_t*)calloc(frame_count,sizeof(struct frame_t));
  for ( i = 0 ; i < frame_count ; i++ )
  { if ( binary_flag )
      read_file_header_binary(filenames[i],i);
    else
      read_file_header_ascii(filenames[i],i);
  }
  
  // See what maximum memory need is for any frame, so we can
  // allocate same size for all frames, so they are exchangeable
  // if there are more frames than will fit into memory.
  maxmem = 0;
  maxfilesize = 0;
  maxfacets = 0;
  printf("\n");
  for ( i = 0 ; i < frame_count ; i++ )
  { struct frame_t *f = frames + i;
    int this_mem = 3*f->vcount*sizeof(float) + 3*f->vcount*sizeof(float) + 
		7*f->fcount*sizeof(int) + 3*f->ecount*sizeof(int);
    if ( this_mem > maxmem )
		maxmem = this_mem;
    if ( f->fcount > maxfacets ) 
        maxfacets = f->fcount;
	if ( f->filelength > maxfilesize ) 
		maxfilesize = f->filelength;

    if ( (i % 10) == 9 )
    { printf("\rReading frame %d    ",i+1);
      fflush(stdout);
    }
      
  }
  printf("\n");
  
  // for reading in whole file at once
  file_buffer = calloc(maxfilesize,1);

  // See how many frames we want to allocate, and allocate memory.
  free_memory_count = 0.75*available_bytes/(maxmem + 64); // allow malloc overhead
  if ( free_memory_count > frame_count )
	  free_memory_count = frame_count;
  free_memory_frames = (void**)calloc(free_memory_count,sizeof(void*));
  for ( i = 0 ; i < free_memory_count ; i++ )
    free_memory_frames[i] = calloc(maxmem,1);
  depth_sort = (struct depth_s*)calloc(maxfacets+10,sizeof(struct depth_s));

  printf("Allocating %d memory frames for %d frames; total %d MB\n",
	  free_memory_count,frame_count,(maxmem*free_memory_count)/1024/1024);

  // Load first frame
  load_frame(0);
  
  resize();
  
  glutDisplayFunc(draw_screen);
 
  glutMainLoop();

}

//glut32static.lib insists on having __iob, but trying to link
// in libcmt.lib in Visual Studio 2008 gives errors conflicting
// with msvcrt.lib, so have to ignore default library libcmt.lib.
// So put _iob here and hope it works.
char _iob[100000];
